
package HostingMaster::Client;

use strict;
use IO::Socket;
use Digest::MD5 qw(md5_base64);
use vars qw(%default);

%default =
    (
     'remoteport' => 31337,
     'host' => 'localhost',
     'username' => $ENV{'USER'},
     'password' => undef,
     'debug' => 0,
     # 'input' => \*STDIN,
     # 'output' => \*STDOUT,
     # 'session' => 'begin'|'end'|undef,
     # 'secret' => undef,
     );


sub new {
    my $class = ref $_[0] ? ref $_[0] : shift;
    my %settings = ref $_[0] ? %{$_[0]} : @_;
    my $self = { %default };
    foreach( keys %settings ) {
	$self->{$_} = $settings{$_};
    }
    return bless $self, $class;
}


sub connect {
    my ( $self, $host, $port ) = @_;
    $host ||= $self->{'host'};
    $port ||= $self->{'remoteport'};

    $self->{'socket'} = 
      IO::Socket::INET->new( 'Proto' => 'tcp',
			     'PeerAddr' => $host,
			     'PeerPort' => $port )
	  or die "Connection setup failed\n";
    return $self->{'socket'};
}


sub prepare {
    my ( $self, @cmd ) = @_;
    my $user = $self->{'username'} || $ENV{'USER'};
    my $pass = $self->{'password'} || '';
    my $sock = $self->{'socket'} || $self->connect();

    $self->{'command'} = "@cmd";
    $self->{'sessionid'} = md5_base64($self->{'secret'})
	if $self->{'secret'};
    foreach( qw( username command session sessionid ) ) {
	print $sock "$_: $self->{$_}\n" if $self->{$_};
    }
    print $sock "\n";

    chomp( my $in = <$sock> );
    my ( $challenge, $seed ) = $in =~/^Challenge:\s+(\S+)\s*(\S*)$/i;
    my $origpass;
    if( $challenge ) {
	$pass ||= $self->{'password_callback'}->($self)
	    if $self->{'password_callback'};

	$pass = crypt $pass, "\$1\$$seed";
	$pass =~ s/.*\$//;
	$origpass = $pass;
	$pass = md5_base64 "$pass$challenge";

	print $sock "Response: $pass\n\n";
	chomp( $in = <$sock> );

	die "$in\n" unless $in =~ /^ok\b/i;
	$in = '';
    }
    if( $self->{'secret'} or $self->{'session'} eq 'begin' ) {
	chomp( $in = <$sock> ) unless $in;
	if( $in =~ /^SessionID:\s+(\S*)$/i ) {
	    my $id = $1;
	    $seed = $self->{'secret'} || $origpass;
	    my $secret = substr( md5_base64( $seed.$id ), 0, 8 );
	    $secret =~ s/\W+/-/g;
	    $self->{'secret'} = $secret;
	    while( $in ) { chomp( $in = <$sock> ) }
	}
    }

    $sock->autoflush(1);

    return $in ? ( $sock, "$in\n" ) : $sock;
}

sub output {
    my ( $self, @output ) = @_;
    if( ref $self->{'output'} ) {
	my $out = $self->{'output'};
	print $out @output;
    } else {
	$self->{'output'} .= join '', @output;
    }
}

sub input {
    my ( $self ) = @_;
    if( ref $self->{'input'} ) {
	my $in = $self->{'input'};
	return <$in>;
    } else {
	$self->{'input'} =~ s/^([^\n]*(\n|\Z))//s;
	return $1;
    }
}

sub mediate {
    my ( $self, $sock, @output ) = @_;

    $sock ||= $self->{'socket'} or die "Missing socket\n";

    $self->{'output'} = '' unless ref $self->{'output'};
    if( my $pid = fork() ) {
	$self->output( @output ) if @output;

	my $c;
	while( !eof($sock) and defined( $c = getc $sock )
	       and ( $c ne "\04" ) ) {
	    $self->output( $c ) unless $c eq '';
	}
	kill 15, $pid;
	waitpid $pid, 0;
    } else {
	while( my $in = $self->input() ) {
	    warn "send to server: $in\n" if $self->{'debug'};
	    print $sock $in;
	}
	exit 0;
    }

    return $sock;
}


sub run {
    my ( $self, @cmd ) = @_;
    $self->connect();
    my ( $sock, @output ) = $self->prepare( @cmd );
    $self->mediate( $sock, @output );
    close $sock;
}


1;
